/**
 * @file      KdpHostApi.cpp
 * @brief     kdp host api source file
 * @version   0.1 - 2019-04-19
 * @copyright (c) 2019 Kneron Inc. All right reserved.
 */

#include <string.h>
#include <unistd.h>
#include <math.h>

#include "KdpHostApi.h"
#include "err_code.h"
#include "MsgMgmt.h"
#include "KnLog.h"
#include "KnUtil.h"
#include "UARTDev.h"

#include "USBDev_Manage.h"

#include <libusb-1.0/libusb.h>

//#include <thread>
#include <vector>
using namespace std;

#define FLOAT_ZERO_VAL 0.000001
#define MAX_THRESH_VAL 1.0f
#define MIN_THRESH_VAL 0.0f

MsgMgmt*                _p_mgr = NULL; //msg manager
AsyncMsgHandler*        _async_hdlr = NULL;
SyncMsgHandler*         _sync_hdlr = NULL;

user_usb_devices_info user_usb_devices_info_table[MAX_COM_DEV];

static bool is_valid_dev(int dev_idx) {
    if(!_p_mgr) return false;
    return _p_mgr->is_valid_dev(dev_idx);
}

/* Checking if the image dimension is valid.
 * There is preprocessing limitation of image width for KL520 NPU.
 * NIR8: width must be multiple of 4; RGB565/YCBCR422: width must be multiple of 2.
 */
static bool is_valid_input(uint32_t image_format, int32_t image_col, int32_t image_row) {
    bool is_valid = true;

    // bypass preprocessing: BIT19
    if (image_format & 0x80000) {
        return true;
    }

    if ((image_col == 0) || (image_row == 0)) {
        return false;
    }

    switch (image_format & 0x00F0) {
    case NPU_FORMAT_RGB565:
        if (image_col % 2) is_valid = false;
        break;
    case NPU_FORMAT_YCBCR422:
        if (image_col % 2) is_valid = false;
        break;
    case NPU_FORMAT_NIR:
        if (image_col % 4) is_valid = false;
        break;
    default:
        break;
    }
    //printf("%x %d %d\n", image_format, image_col, is_valid);
    return is_valid;
}

//init the host lib
int kdp_lib_init() {
    //_async_hdlr = new AsyncMsgHandler();
    libusb_init(NULL); //init usb lib
    _sync_hdlr = new SyncMsgHandler();
    _p_mgr = new MsgMgmt(_async_hdlr, _sync_hdlr);

    KnPrint(KN_LOG_DBG, "msg handler and msg mgr created.\n");
    return 0;
}

//start the host lib
int kdp_lib_start() {
    if(!_p_mgr) return -1;
    //start receiver, sender thread

    _p_mgr->start();

    //_p_mgr->start_sender();

    //start msg handling thread
    _p_mgr->start_handler();
    KnPrint(KN_LOG_DBG, "handler thread and receiver thread started.\n");
    return 0;
}

//init the lib log
//dir: directory of the log file
//name: file name
int kdp_init_log(const char* dir, const char* name) {
    string dirs(dir);
    string names(name);
    KnLogInit(dirs, names);
    KnSetLogLevel(KN_LOG_STOP);

    KnPrint(KN_LOG_DBG, "\n\n\n************----**********\n");
    KnPrint(KN_LOG_DBG, " -- kdp host lib restarted -- \n");
    KnPrint(KN_LOG_DBG, "************----**********\n\n\n");

    KnPrint(KN_LOG_DBG, "init host lib log to %s%s.\n", dir, name);
    return 0;
}



/* get the total number of current usb devices(KL520) */
int kdp_get_all_user_devices(struct user_usb_devices_info *user_devices_tab)
{
    libusb_device ** devs;
    libusb_device *dev;

    struct libusb_device_descriptor desc;

    int i = 0;
    int dev_num = 0;
    int rv = -1;
    ssize_t cnt = libusb_get_device_list(NULL, &devs);
    if(cnt < 0) {
        KnPrint(KN_LOG_ERR, "get usb device list failed, error code:%d.\n", cnt);
        return (int) cnt;
    }

    while( (dev = devs[i++]) != NULL) {
        rv = libusb_get_device_descriptor(dev, &desc);
        if(rv != 0) {
            KnPrint(KN_LOG_ERR, "get usb device descriptor failed, error code:%d.\n", rv);
            continue;
        }

         if(desc.idProduct == PRODUCT_ID && desc.idVendor == VENDOR_ID) {
            KnPrint(KN_LOG_DBG, "found correct usb dev vid:%d, pid:%d,%d.\n", 
                VENDOR_ID, PRODUCT_ID, desc.iSerialNumber);
            (user_devices_tab+dev_num)->user_device_status = STA_DISCOVERED;
            if(desc.iSerialNumber != 0)
            {
                (user_devices_tab+dev_num)->user_device_iSerialNumber = desc.iSerialNumber;
                (user_devices_tab+dev_num)->user_device_iManufacturer = desc.iManufacturer;
                (user_devices_tab+dev_num)->user_device_iProduct      = desc.iProduct;
            }
            (user_devices_tab+dev_num)->puser_device = dev;
            dev_num ++;
            //break; 
        }
    }

    return dev_num;

}

int user_dev_idxs[MAX_COM_DEV]={0} ;
USBDEV_Mgmt *USBDev_Mgmt = NULL;
#define AUTO_ASSIGN_FLAG 1
//add device to lib
//currently only UART and USB are supported.
//name: device name
int kdp_add_dev(int type, const char* name) {
    if(type == KDP_UART_DEV) {

#ifdef __linux__
        string u_name(name);
        BaseDev* pDev = new UARTDev(u_name);
        if(pDev->get_dev_fd() < 0) {
            KnPrint(KN_LOG_ERR, "dev:%s could not be opened.\n", name);
            delete pDev;
            return -1;
        }

        pDev->set_baudrate(115200);
        pDev->set_parity();

        KnPrint(KN_LOG_DBG, "adding dev for:%s...\n", name);
        int ret = _p_mgr->add_device(pDev);
        if(ret < 0) delete pDev;
        return ret;
#elif _WIN32
        return 0;
#endif
    } else if (type == KDP_USB_DEV) {

            //int cnt = 0;
            int ret = -1;

                     
            int dev_cnt = 0;

            #define DME_MOBILNET_SN "22312805"
            #define DME_AGE_GENDER_SN "55667788"

           USBDev_Mgmt =  new USBDEV_Mgmt();
            memset(user_dev_idxs, -1, MAX_COM_DEV*sizeof(int));
            #if AUTO_ASSIGN_FLAG
            USBDev_Mgmt->kdp_add_all_usb_devs(user_dev_idxs, &dev_cnt);
            for(int i = 0;i <dev_cnt; i++ )
                 KnPrint(KN_LOG_DBG,"dev index[%d] information: \n",user_dev_idxs[i]);
            //USBDev_Mgmt->kdp_remove_usb_dev(user_dev_idxs[0]);
            //dev_cnt = USBDev_Mgmt->kdp_get_devs_num();
            #else
                ret = USBDev_Mgmt->kdp_discovery_usb_dev(user_usb_devices_info_table,&dev_cnt);  
                KnPrint(KN_LOG_DBG,"\n\n device count:%d \n\n",dev_cnt);
                if(dev_cnt > 0)
                {
                    for(int i =0;i< dev_cnt; i++)
                    {
                        KnPrint(KN_LOG_DBG,"dev[%d] information: \n",i);
                        KnPrint(KN_LOG_DBG,"    status: %d\n",user_usb_devices_info_table[i].user_device_status);
                        KnPrint(KN_LOG_DBG, "   path:[%d:%d]...\n",user_usb_devices_info_table[i].bus_num,user_usb_devices_info_table[i].port_num);
                        if(user_usb_devices_info_table[i].user_device_iSerialNumber != 0)
                        {
                            KnPrint(KN_LOG_DBG, "   serial number:%s...\n", user_usb_devices_info_table[i].SerialNumber);
                            if(strcmp(DME_MOBILNET_SN, (char *)user_usb_devices_info_table[i].SerialNumber) == 0)
                            {
                                
                                user_dev_idxs[i] = 2;
                            }
                        }
                        else
                        {
                            user_dev_idxs[i] = 1;
                        }

                        if(user_usb_devices_info_table[i].user_device_iManufacturer != 0)
                        {
                            KnPrint(KN_LOG_DBG, "   Manufacturer:%s...\n", user_usb_devices_info_table[i].Manufacturer_name);
                        }

                        if(user_usb_devices_info_table[i].user_device_iProduct != 0)
                        {
                            KnPrint(KN_LOG_DBG, "   product name:%s...\n", user_usb_devices_info_table[i].Product_name);
                        }
                        
                        ret = USBDev_Mgmt->kdp_add_usb_dev(&user_usb_devices_info_table[i], &user_dev_idxs[i], false);
                        KnPrint(KN_LOG_DBG,"add usb ret is %d\n",ret);

                    }

                }
            #endif

            if(dev_cnt > 1)
            {
                ret = dev_cnt;
            }else if(dev_cnt == 1)
            {
                ret = 0;
            }

            return ret;
    } else {
        KnPrint(KN_LOG_DBG, "No identifiable device\n");
        return -1;
    }
}

//free the resources
int kdp_lib_de_init() {
    if(_p_mgr) {
        _p_mgr->stop_receiver();
        //_p_mgr->stop_sender();
        _p_mgr->stop_handler();

        for(int i = 0; i < 5000; i++) {
            usleep(1000);
            if(_p_mgr->is_receiver_running() == false &&
              //_p_mgr->is_sender_running() == false &&
                _p_mgr->is_handler_running() == false) break;
        }
    }

    usleep(1000);
    KnPrint(KN_LOG_DBG, "threads stopped, starting deconstructing objs...\n");
    if(_p_mgr) delete _p_mgr;
    if(_async_hdlr) delete _async_hdlr;
    if(_sync_hdlr) delete _sync_hdlr;

    libusb_exit(NULL);
    //close log
    KnLogClose();
    return 0;
}

int kdp_sfid_start_sync(int dev_idx, uint32_t* rsp_code, uint32_t* img_size, float thresh,
                uint16_t width, uint16_t height, uint32_t format) {
    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_sfid_start(dev_idx, rsp_code, img_size, thresh,
                width, height, format);
}

int kdp_sfid_start_lw3d_sync(int dev_idx, uint32_t* rsp_code, uint32_t* rgb_size, \
            uint32_t* nir_size, float rgb_thresh, float nir_thresh, uint16_t rgb_width, \
            uint16_t rgb_height, uint16_t nir_width, uint16_t nir_height, \
            uint32_t rgb_fmt, uint32_t nir_fmt) {
    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_sfid_start_lw3d(dev_idx, rsp_code,
            rgb_size, nir_size, rgb_thresh, nir_thresh, rgb_width, rgb_height,
            nir_width, nir_height, rgb_fmt, nir_fmt);
}

int kdp_sfid_new_user_sync(int dev_idx, uint16_t usr_id, uint16_t img_idx) {
    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_sfid_new_user(dev_idx, usr_id, img_idx);
}

int kdp_sfid_register_sync(int dev_idx, uint32_t usr_id, uint32_t* rsp_code) {
    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_sfid_register(dev_idx, usr_id, rsp_code);
}

int kdp_sfid_del_db_sync(int dev_idx, uint32_t usr_id, uint32_t* rsp_code) {
    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_sfid_del_db(dev_idx, usr_id, rsp_code);
}

int kdp_sfid_send_img_gen(int dev_idx, char* img_buf, int buf_len, uint32_t* rsp_code, \
        uint16_t* user_id, uint16_t* mode, uint16_t* mask, char* res) {
    if(!_sync_hdlr) return -1;

    return _sync_hdlr->kdp_sfid_send_img_gen(dev_idx, img_buf, buf_len, rsp_code, \
                user_id, mode, mask, res);
}

int kdp_sfid_send_lw3d_img_gen(int dev_idx, char* img_buf, int buf_len, \
        char* img_buf_nir, int buf_len_nir, uint32_t* rsp_code, \
        uint16_t* user_id, uint16_t* mode, uint16_t* mask, char* res) {
    if(!_sync_hdlr) return -1;

    return _sync_hdlr->kdp_sfid_send_lw3d_gen(dev_idx, img_buf, buf_len, img_buf_nir, \
                buf_len_nir, rsp_code, user_id, mode, mask, res);
}

// -------------------system operation APIs
int kdp_reset_sys(int dev_idx, uint32_t reset_mode) {
    KnPrint(KN_LOG_DBG, "\n\ncalling reset sys...:%d,%x.\n", dev_idx, reset_mode);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_reset_sys_sync(dev_idx, reset_mode);
}

int kdp_report_sys_status(int dev_idx, uint32_t* sfirmware_id, uint32_t* sbuild_id,
            uint16_t* sys_status, uint16_t* app_status, uint32_t* nfirmware_id, uint32_t* nbuild_id) {
    KnPrint(KN_LOG_DBG, "\n\ncalling report sys status...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_report_sys_status_sync(dev_idx, sfirmware_id, \
                sbuild_id, sys_status, app_status, nfirmware_id, nbuild_id);
}

int kdp_get_kn_number(int dev_idx, uint32_t *kn_num) {
    KnPrint(KN_LOG_DBG, "\n\ncalling kdp_get_kn_number: %d\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_get_kn_number(dev_idx, kn_num);
}

int kdp_get_model_info(int dev_idx, int from_ddr, char *data_buf) {
    KnPrint(KN_LOG_DBG, "\n\ncalling kdp_get_model_info: %d\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_get_model_info(dev_idx, from_ddr, data_buf);
}

int kdp_update_fw(int dev_idx, uint32_t* module_id, char* img_buf, int buf_len) {
    KnPrint(KN_LOG_DBG, "\n\ncalling updating firmware...:%d, %d.\n", dev_idx, *module_id);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_update_fw_sync(dev_idx, module_id, img_buf, buf_len);
}

int kdp_update_model(int dev_idx, uint32_t* model_id, uint32_t model_size, \
                    char* img_buf, int buf_len) {
    KnPrint(KN_LOG_DBG, "\n\ncalling updating model...:%d, %d.\n", dev_idx, *model_id);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_update_model_sync(dev_idx, model_id, \
                model_size, img_buf, buf_len);
}

int kdp_update_spl(int dev_idx, uint32_t mode, uint16_t auth_type, uint16_t size, \
            uint8_t* auth_key, char* spl_buf, uint32_t* rsp_code, uint32_t* spl_id, \
            uint32_t* spl_build) {
    KnPrint(KN_LOG_DBG, "\n\ncalling updating spl...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_update_spl_sync(dev_idx, mode, auth_type, size, \
            auth_key, spl_buf, rsp_code, spl_id, spl_build);
}

//-------------------------------app level APIs

int kdp_start_sfid_mode(int dev_idx, uint32_t* img_size, float thresh,
            uint16_t width, uint16_t height, uint32_t format) {
    KnPrint(KN_LOG_DBG, "\n\ncalling start verify mode...:%d.\n", dev_idx);

    if (!is_valid_input(format, width, height)) {
        KnPrint(KN_LOG_ERR, "the image with unacceptable dimension could not be processed by KL520 NPU.\n");
        return -2;
    }

    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }
    if(!img_size) {
        KnPrint(KN_LOG_ERR, "img size is null.\n");
        return -1;
    }

    if(fabsf(thresh) < FLOAT_ZERO_VAL) {
        //0.0f
    } else {
        if(thresh > MAX_THRESH_VAL || thresh < MIN_THRESH_VAL) {
            KnPrint(KN_LOG_ERR, "threshold val is invalid, range:%5.2f ~ %5.2f.\n",
                    MIN_THRESH_VAL, MAX_THRESH_VAL);
            return -1;
        }
    }

    if(width == 0) width = 640;
    if(height == 0) height = 480;

    uint32_t rsp_code = 0;
    int ret = kdp_sfid_start_sync(dev_idx, &rsp_code, img_size,
                thresh, width, height, format);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "start verify mode failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "start verify mode failed with error:%d.\n", rsp_code);
        return rsp_code;
    }
    return 0;
}

int kdp_verify_user_id_generic(int dev_idx, uint16_t* user_id,
                               char* img_buf, int buf_len, uint16_t* mask, char* res)
{
    KnPrint(KN_LOG_DBG, "\n\ncalling verify user id...:%d,%d.\n", dev_idx, buf_len);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!user_id || !img_buf || buf_len <= 0) {
        KnPrint(KN_LOG_ERR, "verify user id input parameter is invalid.\n");
        return -1;
    }

    if( mask && (*mask & 0x0f) ) {
        if(res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for res flag, result.\n");
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    uint16_t mode = 1;
    int ret = kdp_sfid_send_img_gen(dev_idx, img_buf, buf_len, &rsp_code, user_id, &mode,
                mask, res);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "verify user id failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "verify user id failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    if(mode != 0) {
        KnPrint(KN_LOG_ERR, "device is not in user verify mode.\n");
        return -1;
    }

    KnPrint(KN_LOG_DBG, "verify user id completed, user id:%d.\n", *user_id);
    if(*user_id < 1 || *user_id > MAX_USER) {
        KnPrint(KN_LOG_ERR, "invalid user id received:%d.\n", *user_id);
        return -1;
    }
    return 0;
}

int kdp_start_reg_user_mode(int dev_idx, uint16_t usr_id, uint16_t img_idx) {
    KnPrint(KN_LOG_DBG, "\n\ncalling start reg user mode...:%d,%d.\n", usr_id, img_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(usr_id < 1 || usr_id > MAX_USER || img_idx < 1 || img_idx > MAX_FID) {
        KnPrint(KN_LOG_ERR, "new user input parameter is invalid.\n");
        return -1;
    }

    int ret =  kdp_sfid_new_user_sync(dev_idx, usr_id, img_idx);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "start reg mode failed with error:%d.\n", ret);
        return -1;
    }

    return 0;
}

uint16_t kdp_get_res_mask(bool fd, bool lm, bool fr, bool liveness)
{
    return ((fd << 0) + (lm << 1) + (fr << 2) + (liveness << 3));
}

uint32_t kdp_get_res_size(bool fd, bool lm, bool fr, bool liveness)
{
    uint32_t size = 0;

    size += fd ? FD_RES_LENGTH : 0;
    size += lm ? LM_RES_LENGTH : 0;
    size += fr ? FR_RES_LENGTH : 0;
    size += liveness ? LV_RES_LENGTH : 0;

    return size;
}

int kdp_extract_feature_generic(int dev_idx,
                                char* img_buf, int buf_len, uint16_t* mask, char* res)
{
    KnPrint(KN_LOG_DBG, "\n\ncalling extract feature generic..:%d,%d.\n", dev_idx, buf_len);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!img_buf || buf_len <= 0) {
        KnPrint(KN_LOG_ERR, "extract feature input parameter is invalid.\n");
        return -1;
    }

    if( mask && (*mask & 0x0f) ) {
        if(res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for res flag, result.\n");
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    uint16_t img_idx = 0;
    uint16_t mode = 0;
    int ret = kdp_sfid_send_img_gen(dev_idx, img_buf, buf_len, &rsp_code, &img_idx, &mode,
                mask, res);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "extract feature failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "extract feature failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    if(mode == 0) {
        KnPrint(KN_LOG_ERR, "device is not in user register mode.\n");
        return -1;
    }

    KnPrint(KN_LOG_DBG, "extract feature completed, img index:%d.\n", img_idx);
    if(img_idx < 1 || img_idx > 5) {
        KnPrint(KN_LOG_ERR, "invalid image received:%d.\n", img_idx);
        return -1;
    }
    return 0;
}

int kdp_register_user(int dev_idx, uint32_t user_id) {
    KnPrint(KN_LOG_DBG, "\n\ncalling register user...:%d,%d.\n", dev_idx, user_id);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(user_id < 1 || user_id > MAX_USER) {
        KnPrint(KN_LOG_ERR, "reg user input parameter is invalid.\n");
        return -1;
    }

    uint32_t rsp_code = 0;
    int ret = kdp_sfid_register_sync(dev_idx, user_id, &rsp_code);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "reg user failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "reg user failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    KnPrint(KN_LOG_DBG, "reg user completed, user id:%d.\n", user_id);
    return 0;
}

int kdp_remove_user(int dev_idx, uint32_t user_id) {
    KnPrint(KN_LOG_DBG, "\n\ncalling remove user...:%d,%d.\n", dev_idx, user_id);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }
    if(user_id < 0 || user_id > MAX_USER) {
        KnPrint(KN_LOG_ERR, "del user input parameter is invalid.\n");
        return -1;
    }

    uint32_t rsp_code = 0;
    int ret = kdp_sfid_del_db_sync(dev_idx, user_id, &rsp_code);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "remove all user failed for dev:%d, with error:%d.\n", dev_idx, ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "remove all user failed for dev:%d, with error:%d.\n", dev_idx, rsp_code);
        return rsp_code;
    }
    return 0;
}

int kdp_start_lw3d_mode(int dev_idx,
                        uint32_t* rgb_size, uint32_t* nir_size,
                        float rgb_thresh, float nir_thresh,
                        uint16_t rgb_width, uint16_t rgb_height,
                        uint16_t nir_width, uint16_t nir_height,
                        uint32_t rgb_fmt, uint32_t nir_fmt) {

    KnPrint(KN_LOG_DBG, "\n\ncalling start lw3d verify mode...:%d.\n", dev_idx);

    if ((!is_valid_input(rgb_fmt, rgb_width, rgb_height)) || (!is_valid_input(nir_fmt, nir_width, nir_height))) {
        KnPrint(KN_LOG_ERR, "the image with unacceptable dimension could not be processed by KL520 NPU.\n");
        return -2;
    }

    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }
    if(!rgb_size || !nir_size) {
        KnPrint(KN_LOG_ERR, "img size is null.\n");
        return -1;
    }

    if(fabsf(rgb_thresh) < FLOAT_ZERO_VAL) {
        //0.0f
    } else {
        if(rgb_thresh > MAX_THRESH_VAL || rgb_thresh < MIN_THRESH_VAL) {
            KnPrint(KN_LOG_ERR, "rgb threshold val is invalid, range:%5.2f ~ %5.2f.\n",
                    MIN_THRESH_VAL, MAX_THRESH_VAL);
            return -1;
        }
    }

    if(fabsf(nir_thresh) < FLOAT_ZERO_VAL) {
        //0.0f
    } else {
        if(nir_thresh > MAX_THRESH_VAL || nir_thresh < MIN_THRESH_VAL) {
            KnPrint(KN_LOG_ERR, "nir threshold val is invalid, range:%5.2f ~ %5.2f.\n",
                    MIN_THRESH_VAL, MAX_THRESH_VAL);
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    int ret = kdp_sfid_start_lw3d_sync(dev_idx, &rsp_code, rgb_size,
                nir_size, rgb_thresh, nir_thresh, rgb_width, rgb_height,
                nir_width, nir_height, rgb_fmt, nir_fmt);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "start lw3d verify mode failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "start lw3d verify mode failed with error:%d.\n", rsp_code);
        return rsp_code;
    }
    return 0;
}

int kdp_verify_lw3D_image_generic(int dev_idx, uint16_t* user_id, char* img_buf, int buf_len,
                                  char* img_buf_nir, int buf_len_nir, uint16_t* mask, char* res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling verify user id lw3d generic...:%d,%d.\n", dev_idx, buf_len);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!user_id || !img_buf || buf_len <= 0) {
        KnPrint(KN_LOG_ERR, "verify user id lw3d rgb input parameter is invalid.\n");
        return -1;
    }

    if(!img_buf_nir || buf_len_nir <= 0) {
        KnPrint(KN_LOG_ERR, "verify user id lw3d nir input parameter is invalid.\n");
        return -1;
    }

    if( mask && (*mask & 0x0f) ) {
        if(res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for lw3d res flag, result.\n");
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    uint16_t mode = 1;
    int ret = kdp_sfid_send_lw3d_img_gen(dev_idx, img_buf, buf_len, img_buf_nir, buf_len_nir,
                                         &rsp_code, user_id, &mode, mask, res);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "verify user id lw3d failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "verify user id ls3d failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    KnPrint(KN_LOG_DBG, "verify user id lw3d completed, user id:%d.\n", *user_id);
    if(*user_id < 1 || *user_id > MAX_USER) {
        KnPrint(KN_LOG_ERR, "invalid user id received:%d.\n", *user_id);
        return -1;
    }
    return 0;
}

uint16_t kdp_get_lw3D_res_mask(bool fd, bool lm, bool fr,
                               bool nir_fd, bool nir_lm, bool nir_fr, bool liveness)
{
    return ((fd << 0) + (lm << 1) + (fr << 2) +
            (nir_fd << 4) + (nir_lm << 5) + (nir_fr << 6) + (liveness << 8));
}

uint32_t kdp_get_lw3D_res_size(bool fd, bool lm, bool fr, bool nir_fd, bool nir_lm, bool nir_fr, bool liveness)
{
    uint32_t size = 0;

    size += fd ? FD_RES_LENGTH : 0;
    size += lm ? LM_RES_LENGTH : 0;
    size += fr ? FR_RES_LENGTH : 0;
    size += nir_fd ? FD_RES_LENGTH : 0;
    size += nir_lm ? LM_RES_LENGTH : 0;
    size += nir_fr ? FR_RES_LENGTH : 0;
    size += liveness ? LV_RES_LENGTH : 0;

    return size;
}

int kdp_extract_lw3D_feature_generic(int dev_idx, char* img_buf, int buf_len,
                                     char* img_buf_nir, int buf_len_nir, uint16_t* mask, char* res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling extract lw3d feature generic...:%d,%d.\n", dev_idx, buf_len);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!img_buf || buf_len <= 0) {
        KnPrint(KN_LOG_ERR, "extract feature lw3d rgb input parameter is invalid.\n");
        return -1;
    }

    if(!img_buf_nir || buf_len_nir <= 0) {
        KnPrint(KN_LOG_ERR, "extract feature lw3d nir input parameter is invalid.\n");
        return -1;
    }

    if( mask && (*mask & 0x0f) ) {
        if(res == NULL) {
            KnPrint(KN_LOG_ERR, "invalid input arguments for lw3d res flag, result.\n");
            return -1;
        }
    }

    uint32_t rsp_code = 0;
    uint16_t mode = 1;
    uint16_t user_id = 0;
    int ret = kdp_sfid_send_lw3d_img_gen(dev_idx, img_buf, buf_len, img_buf_nir, buf_len_nir,
                    &rsp_code, &user_id, &mode, mask, res);
    if(ret < 0) {
        KnPrint(KN_LOG_ERR, "extract feature lw3d failed with error:%d.\n", ret);
        return -1;
    }

    if(rsp_code != 0) {
        KnPrint(KN_LOG_ERR, "extract feature lw3d failed with error:%d.\n", rsp_code);
        return rsp_code;
    }

    KnPrint(KN_LOG_DBG, "extract feature lw3d completed, user id:%d.\n", user_id);
    if(user_id < 1 || user_id > 5) {
        KnPrint(KN_LOG_ERR, "invalid user id received:%d.\n", user_id);
        return -1;
    }
    return 0;
}

float kdp_fm_compare(float *user_fm_a, float *user_fm_b, size_t fm_size)
{
    //**************
    // calcualte match
    //**************
    float sum_xy = 0.;

    if ((user_fm_a == NULL) || (user_fm_b == NULL) || (fm_size <= 0)) {
        printf("parameter error \n");
        return -1;
    }

    for (size_t i = 0; i < fm_size; i++) {
        sum_xy += powf((user_fm_a[i] - user_fm_b[i]), 2);
    }

    return sqrtf(sum_xy);
}

//-------------------------------ISI APIs
int kdp_start_isi_mode(int dev_idx, uint32_t app_id, uint32_t rt_size, uint16_t width, uint16_t height, uint32_t format, uint32_t* rsp_code, uint32_t* buf_size) {
    KnPrint(KN_LOG_DBG, "\n\ncalling start isi mode...:%d.\n", dev_idx);

    if (!is_valid_input(format, width, height)) {
        KnPrint(KN_LOG_ERR, "the image with unacceptable dimension could not be processed by KL520 NPU.\n");
        return -2;
    }

    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_isi_start(dev_idx, app_id, rt_size, width, height, format, rsp_code, buf_size);
}

int kdp_start_isi_mode_ext(int dev_idx, char* isi_cfg, int cfg_size, uint32_t* rsp_code, uint32_t* buf_size) {
    KnPrint(KN_LOG_DBG, "\n\ncalling start isi mode...:%d.\n", dev_idx);

    struct kdp_isi_cfg_s *cfg = (struct kdp_isi_cfg_s *)isi_cfg;
    if (!is_valid_input(cfg->image_format, cfg->image_col, cfg->image_row)) {
        KnPrint(KN_LOG_ERR, "the image with unacceptable dimension could not be processed by KL520 NPU.\n");
        return -2;
    }

    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_isi_start_ext(dev_idx, isi_cfg, cfg_size, rsp_code, buf_size);
}

int kdp_isi_inference(int dev_idx, char* img_buf, int buf_len, uint32_t img_id, \
            uint32_t* rsp_code, uint32_t* img_available) {
    KnPrint(KN_LOG_DBG, "\n\ncalling isi inference...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_isi_inference(dev_idx, img_buf, buf_len, img_id, rsp_code, img_available);
}

int kdp_isi_retrieve_res(int dev_idx, uint32_t img_id, uint32_t* rsp_code, uint32_t* r_size, char* r_data) {
    KnPrint(KN_LOG_DBG, "\n\ncalling isi retrieve results...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_isi_retrieve_res(dev_idx, img_id, rsp_code, r_size, r_data);
}

//-------------------------------DME APIs
int kdp_start_dme(int dev_idx, uint32_t model_size, char* data, int dat_size, \
                  uint32_t* ret_size, char* img_buf, int buf_len) {
    KnPrint(KN_LOG_DBG, "\n\ncalling starting dme mode...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_start_dme(dev_idx, model_size, data, dat_size,
                    ret_size, img_buf, buf_len);
}

int kdp_dme_configure(int dev_idx, char* data, int dat_size, uint32_t* ret_model_id) {
    KnPrint(KN_LOG_DBG, "\n\ncalling dme config...:%d.\n", dev_idx);

    struct kdp_dme_cfg_s *dme_cfg = (struct kdp_dme_cfg_s *)data;
    if (!is_valid_input(dme_cfg->image_format, dme_cfg->image_col, dme_cfg->image_row)) {
        KnPrint(KN_LOG_ERR, "the image with unacceptable dimension could not be processed by KL520 NPU.\n");
        return -2;
    }

    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_dme_configure(dev_idx, data, dat_size, ret_model_id);
}

int kdp_dme_inference(int dev_idx, char* img_buf, int buf_len, uint32_t* inf_size, \
            bool* res_flag, char* inf_res, uint16_t mode, uint16_t model_id) {
    KnPrint(KN_LOG_DBG, "\n\ncalling dme inference...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_dme_inference(dev_idx, img_buf, buf_len, inf_size, res_flag, inf_res, mode, model_id);
}

int kdp_dme_retrieve_res(int dev_idx, uint32_t addr, int len, char* inf_res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling dme retrieve results...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_dme_retrieve_res(dev_idx, addr, len, inf_res);
}

int kdp_dme_get_status(int dev_idx, uint16_t *ssid, uint16_t *status, uint32_t* inf_size, char* inf_res) {
    KnPrint(KN_LOG_DBG, "\n\ncalling dme get status...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_dme_get_status(dev_idx, ssid, status, inf_size, inf_res);
}

int kdp_end_dme(int dev_idx) {
    KnPrint(KN_LOG_DBG, "\n\ncalling ending dme...:%d.\n", dev_idx);
    if(!is_valid_dev(dev_idx)) {
        KnPrint(KN_LOG_ERR, "dev idx is invalid.\n");
        return -1;
    }

    if(!_sync_hdlr) return -1;
    return _sync_hdlr->kdp_reset_sys_sync(dev_idx, 255);
}

void error_code_2_str(int code, char* buf, int len)
{
    memset(buf, 0, len);

    switch(code) {
    case UNKNOWN_ERR:
        snprintf(buf, len-1, "General error");
        break;
    case MSG_APP_NOT_LOADED:
        snprintf(buf, len-1, "SFID application not loaded");
        break;
    case MSG_APP_BAD_INDEX:
        snprintf(buf, len-1, "Invalid User ID or image index specified");
        break;
    case MSG_APP_UID_EXIST:
        snprintf(buf, len-1, "User ID already exists");
        break;
    case MSG_APP_UID_DIFF:
        snprintf(buf, len-1, "User ID changed from the New User session");
        break;
    case MSG_APP_IDX_FIRST:
        snprintf(buf, len-1, "User ID changed from the New User session");
        break;
    case MSG_APP_IDX_MISSING:
        snprintf(buf, len-1, "No feature maps from New User session found");
        break;
    case MSG_APP_DB_NO_USER:
        snprintf(buf, len-1, "User ID not found in DB");
        break;
    case MSG_APP_DB_FAIL:
        snprintf(buf, len-1, "DB delete failure");
        break;
    case MSG_APP_DB_NO_MATCH:
        snprintf(buf, len-1, "No match in User DB");
        break;
    case MSG_APP_BAD_IMAGE:
        snprintf(buf, len-1, "No successful inference (FD-FR)");
        break;
    default:
        snprintf(buf, len-1, "error code [%d] not found", code);
        break;
    }
    return;
}
