/**
 * @file      USBDev.cpp
 * @brief     implementation file for USB Dev class
 * @version   0.1 - 2019-04-28
 * @copyright (c) 2019 Kneron Inc. All right reserved.
 */

#include "USBDev.h"
#include "KnLog.h"
#include "KdpHostApi.h"
#include "UARTPkt.h"
#include "KnUtil.h"
#include "KnMsg.h"

#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include <unistd.h>
#include <stdlib.h>

#define MAX_RETRY_NUM    2
#define BULK_SIZE_USB    512
#define LIB_USB_TIMEOUT  5000

int USBDev::_inst_cnt = 0;

USBDev::USBDev(int vid, int pid) {
    m_pkneron_device_h = NULL;
    p_kneron_dev = NULL;
    vendor_id = vid;
    product_id = pid;
    dev_type = KDP_USB_DEV;
    p_dev_list = NULL;

    _inst_cnt++;
    dev_fd = USB_FAKE_FD_START + _inst_cnt;

    memset(serialnumber, 0, 33);
}


USBDev::~USBDev() {
    if (NULL != m_pkneron_device_h) {

        KnPrint(KN_LOG_DBG, "reseting usb device before exiting,%08x.\n",m_pkneron_device_h);
        libusb_reset_device(m_pkneron_device_h);
        KnPrint(KN_LOG_DBG, "usb device has been reset.\n");

        libusb_release_interface(m_pkneron_device_h, 0);
        libusb_close(m_pkneron_device_h);
        m_pkneron_device_h = NULL;
    }
    if(p_dev_list) {
        libusb_free_device_list(p_dev_list, 1);
    }

    _inst_cnt--;
}

int USBDev::u_init_port(struct libusb_device* device)
{
    p_kneron_dev = device;
    return 0;

}

int USBDev::u_get_device_list()
 {
    libusb_device** devs;
    size_t cnt = libusb_get_device_list(NULL, &devs);
    if(cnt < 0) {
        KnPrint(KN_LOG_ERR, "get usb device list failed, error code:%d.\n", cnt);
        return (int) cnt;
    }

    p_dev_list = devs;
    return cnt;
 }
 struct libusb_device* USBDev::u_find_macth_dev(struct libusb_device** dev_list, uint8_t* serial_number)
 {
    int i = 0;
    int rv = -1;
    struct libusb_device_descriptor desc;
    struct libusb_device* dev;
    libusb_device_handle *dev_handle = NULL;
    unsigned char snum_buf[33] = {0};

    if(dev_list == NULL)
    {
        KnPrint(KN_LOG_ERR, "devices list is null\n");
        return NULL;
    }

  
     while( (dev = dev_list[i++]) != NULL) {
        rv = libusb_get_device_descriptor(dev, &desc);
        if(rv != 0) {
            KnPrint(KN_LOG_ERR, "get usb device descriptor failed, error code:%d.\n", rv);
            continue;
        }

        if(desc.idProduct == product_id && desc.idVendor == vendor_id) {
            KnPrint(KN_LOG_DBG, "found correct usb dev vid:%d, pid:%d,%d.\n", 
                vendor_id, product_id, desc.iSerialNumber);
            
           
            if(strlen((char *)serial_number) == 0){
                if(desc.iSerialNumber == 0)
                { 
                    uint8_t re_usb_bus = libusb_get_bus_number(dev);
                    uint8_t re_usb_port = 0;
                    if(re_usb_bus == usb_bus)
                    {
                        re_usb_port = libusb_get_port_number(dev);
                        if(re_usb_port == usb_port)
                        {
                            KnPrint(KN_LOG_DBG, "[WARN] find a device without serial number,please check it \n");
                            break;
                        }
                    }
                }
                continue;
                   
            }

            int status = libusb_open(dev, &dev_handle);
            if( status != 0 || !dev_handle ) {
                KnPrint(KN_LOG_ERR, "reconnect usb device vid:%d, pid:%d, open failed: [%s].\n",vendor_id, product_id, libusb_error_name(status));
                return NULL;
            }

            if (libusb_get_string_descriptor_ascii(dev_handle, desc.iSerialNumber, snum_buf, 33) >= 0) 
            {
                libusb_close(dev_handle);
                //sscanf((char *)snum_buf, "%x", &snum_tmp);
                KnPrint(KN_LOG_DBG, "find a reconnect usb device serial_number is:%s\n",snum_buf);
                if(strcmp((char *)snum_buf, (char *)serial_number) == 0)
                {
                    break;
                }
            }
            
           
           
            //break; 
        }
    }

    return dev;
 }
int USBDev::u_init_port()
{
   // libusb_device** devs;
    libusb_device* dev;

    ssize_t cnt = u_get_device_list();
    if(cnt < 0) {
        KnPrint(KN_LOG_ERR, "get usb device list failed, error code:%d.\n", cnt);
        return (int) cnt;
    }

    int i = 0;
    struct libusb_device_descriptor desc;

    int rv = -1;
    while( (dev = p_dev_list[i++]) != NULL) {
        rv = libusb_get_device_descriptor(dev, &desc);
        if(rv != 0) {
            KnPrint(KN_LOG_ERR, "get usb device descriptor failed, error code:%d.\n", rv);
            continue;
        }

        if(desc.idProduct == product_id && desc.idVendor == vendor_id) {
            KnPrint(KN_LOG_DBG, "found correct usb dev vid:%d, pid:%d.\n", 
                vendor_id, product_id);
            p_kneron_dev = dev;
            break; 
        }
    }

    if(!p_kneron_dev) {
        KnPrint(KN_LOG_ERR, "no usb device is found, error code:%d.\n", rv);
        return -1;
    } else {
        return 0;
    }
}

int USBDev::u_open() {
    if (NULL == p_kneron_dev) return -1;

    libusb_device_handle * dev_tmp_h = m_pkneron_device_h;

    int status = libusb_open(p_kneron_dev, &m_pkneron_device_h);
    if( status != 0 || !m_pkneron_device_h ) {
        KnPrint(KN_LOG_ERR, "usb device vid:%d, pid:%d, open failed: [%s].\n",
                vendor_id, product_id, libusb_error_name(status));
        return -1;
    }
    else
    {
        // release the old device
        if (NULL != dev_tmp_h) {
            libusb_release_interface(dev_tmp_h, 0);
            libusb_close(dev_tmp_h);
            dev_tmp_h = NULL;
        }
    }

    if (0 != (status =  libusb_set_configuration(m_pkneron_device_h, 1))) {
        KnPrint(KN_LOG_ERR, "error for set usb config: [%s].\n", libusb_error_name(status));
        libusb_close(m_pkneron_device_h);
        m_pkneron_device_h = NULL;
        return -1;
    }

    if (0 != (status = libusb_claim_interface(m_pkneron_device_h, 0))) {
        KnPrint(KN_LOG_ERR, "claim interface failed: [%s].\n", libusb_error_name(status));
        libusb_close(m_pkneron_device_h);
        m_pkneron_device_h = NULL;
        return -1;
    }
    KnPrint(KN_LOG_DBG, "usb dev opened successfully.\n");
    return 0;
}

int USBDev::send(int8_t* msgbuf, int msglen) {
    return send((char*) msgbuf, msglen);
}

int USBDev::send(char* msgbuf, int msglen) {
    int status = 0;

    if (NULL == m_pkneron_device_h)
    {
         KnTsPrint(KN_LOG_DBG, "send failed, m_pkneron_device_h is NULL.\n");
         return -1;
    }

    //need to pack the data
    KnTsPrint(KN_LOG_DBG, "packing payload with uart pkt:%d.\n", msglen);
    UARTPkt pkt((int8_t*)msgbuf, msglen, true, MSG_CRC_FLAG);

    //calling base senf func to send the packed data
    if(!pkt.is_pkt_valid()) {
        KnPrint(KN_LOG_ERR, "packing to UART packet falied.\n");
        return -1;
    }

    kn_dump_msg(pkt.get_pkt_buf(), pkt.get_pkt_len());

    KnTsPrint(KN_LOG_DBG, "kdp msg is packed with uart pkt of size [%d].\n", pkt.get_pkt_len());

    int transfered = 0;
    KnTsPrint(KN_LOG_DBG, "calling usb interrupt send to send data of size [%d]...\n", msglen);
    status = libusb_bulk_transfer(m_pkneron_device_h, EP_OUT_INT,
            (unsigned char*)(pkt.get_pkt_buf()), pkt.get_pkt_len(), &transfered, LIB_USB_TIMEOUT);
    if(status < 0) {
        KnPrint(KN_LOG_ERR, "write usb failed: %d, %s.\n", status, libusb_error_name(status));
        return status;
    }

    check_to_send_zero_byte(transfered, EP_OUT_INT);

    KnTsPrint(KN_LOG_DBG, "usb write sent data :%d.\n", transfered);
    return transfered;
}

int USBDev::send_bulk(char* msgbuf, int msglen, bool send_zero_byte)
{
    int status = 0;

    if (NULL == m_pkneron_device_h)
        return -1;

    int transfered = 0;
    KnTsPrint(KN_LOG_DBG, "calling usb bulk send to send data :%d...\n", msglen);
    status = libusb_bulk_transfer(m_pkneron_device_h, EP_OUT_BULK,
                                  (unsigned char*)msgbuf, msglen, &transfered, LIB_USB_TIMEOUT);
    if(status < 0) {
        KnPrint(KN_LOG_ERR, "write usb failed: %d, %s.\n", status, libusb_error_name(status));
        return status;
    }

    if (send_zero_byte) {
        check_to_send_zero_byte(transfered, EP_OUT_BULK);
    }

    KnTsPrint(KN_LOG_DBG, "usb write sent data :%d.\n", transfered);
    return transfered;
}

/* due to USB protocol, when data len is multiple of BULK_SIZE_USB,
 * we should send a zero-len data to notify no more data */
int USBDev::check_to_send_zero_byte(size_t len, unsigned char endpoint)
{
    if (len == 0 || (len % BULK_SIZE_USB) > 0) {
        return 0;
    }

    unsigned char data[1];
    int tmp = 0;

    KnTsPrint(KN_LOG_DBG, "sending zero byte...\n");

    return libusb_bulk_transfer(m_pkneron_device_h, endpoint, data, 0, &tmp, LIB_USB_TIMEOUT);
}

int USBDev::check_to_receive_zero_byte(size_t len, unsigned char endpoint)
{
    if (len == 0 || (len % BULK_SIZE_USB) > 0) {
        return 0;
    }

    unsigned char data;
    int tmp = 0;

    KnTsPrint(KN_LOG_DBG, "receiving zero byte...\n");

    int ret = libusb_bulk_transfer(m_pkneron_device_h, endpoint, &data, sizeof(unsigned char), &tmp, LIB_USB_TIMEOUT);

    if(ret == 0 && tmp !=0)
        printf("%s() failed, should be ZLP but it is not !!\n", __FUNCTION__);

    return ret;
}

int USBDev::receive(int8_t* msgbuf, int msglen) {
    return receive((char*)msgbuf, msglen);
}

void USBDev::shutdown_dev() {
    KnTsPrint(KN_LOG_DBG, "shutdown usb device...\n");

    if (NULL != m_pkneron_device_h) {
        libusb_release_interface(m_pkneron_device_h, 0);
        libusb_close(m_pkneron_device_h);
        m_pkneron_device_h = NULL;
    }
    if(p_dev_list) {
        libusb_free_device_list(p_dev_list, 1);
        p_dev_list = NULL;
        p_kneron_dev = NULL;
    }
}

void USBDev::restart_dev() {
    KnTsPrint(KN_LOG_DBG, "restarting usb device...\n");

    KnTsPrint(KN_LOG_DBG, "calling init port...\n");
    if(u_init_port() < 0) {
        KnPrint(KN_LOG_ERR, "init port failed for restarting usb.\n");
        return;
    }

    KnTsPrint(KN_LOG_DBG, "calling usb open...\n");
    if(u_open() < 0) {
        KnPrint(KN_LOG_ERR, "usb open failed for restarting usb.\n");
        return;
    }
}


int USBDev::reconnect_usb() {
    //close usb
    KnTsPrint(KN_LOG_DBG, "re connecting usb for no such device error...\n");

    if(p_dev_list) {
        libusb_free_device_list(p_dev_list, 1);
        p_dev_list = NULL;
        p_kneron_dev = NULL;
    }

    usleep(10000);

    KnTsPrint(KN_LOG_DBG, "calling init port...\n");
    /*
    if(u_init_port() < 0) {
        KnPrint(KN_LOG_ERR, "init port failed for reconnecting usb.\n");
        return -1;
    }*/
    if(u_get_device_list()< 0)
    {
        KnPrint(KN_LOG_ERR, "get usb list failed for reconnecting usb.\n");
        return -1;
    }

    p_kneron_dev = u_find_macth_dev(p_dev_list, serialnumber);
    if(p_kneron_dev == NULL)
    {
        if(strlen((char *)serialnumber) > 0)
            KnPrint(KN_LOG_ERR, "get usb devices[%s] failed for reconnecting usb.\n", serialnumber);
        else
        {
            KnPrint(KN_LOG_ERR, "reconnect usb devices failed: cannt find the devices[%d:%d] \n",usb_bus,usb_port);
        }
        return -2;
    }

    KnTsPrint(KN_LOG_DBG, "calling usb open...\n");
    if(u_open() < 0) {
        KnPrint(KN_LOG_ERR, "usb open failed for reconnecting usb.\n");
        return -1;
    }
    return 0;
}

int USBDev::receive(char* msgbuf, int msglen) {
    int ret_val;
    if (NULL == m_pkneron_device_h)
        return -1;

    int transfered = 0;
    KnTsPrint(KN_LOG_DBG, "calling usb interrupt read to receive data :%d...\n", msglen);

    for(int i = 0; i < MAX_RETRY_NUM; i++) {
        ret_val = libusb_bulk_transfer(m_pkneron_device_h, EP_IN_INT ,
                                       (unsigned char*)msgbuf, msglen, &transfered, LIB_USB_TIMEOUT);
        if(ret_val != 0) {
            KnTsPrint(KN_LOG_ERR, "USB EP read failed:%d, %s.\n", ret_val, libusb_error_name(ret_val));
            if(ret_val != LIBUSB_ERROR_NO_DEVICE) {
                return -1;
            } else { //no such device
                int s = reconnect_usb();
                if(s < 0) return -1;
                continue; //retry
            }
        } else {
            check_to_receive_zero_byte(transfered, EP_IN_INT);
            break;
        }
    }

    KnTsPrint(KN_LOG_DBG, "usb read received data :%d.\n", transfered);
    return transfered;
}

int USBDev::receive_bulk(char* msgbuf, int msglen) {
    int ret_val;
    if (NULL == m_pkneron_device_h)
        return -1;

    int transfered = 0;
    KnTsPrint(KN_LOG_DBG, "calling usb bulk read to receive data :%d...\n", msglen);

    for(int i = 0; i < MAX_RETRY_NUM; i++) {
        ret_val = libusb_bulk_transfer(m_pkneron_device_h, EP_IN_BULK ,
                                       (unsigned char*)msgbuf, msglen, &transfered, LIB_USB_TIMEOUT);
        if (ret_val != 0) {
            KnTsPrint(KN_LOG_ERR, "USB EP read failed:%d, %s.\n", ret_val, libusb_error_name(ret_val));
            if(ret_val != LIBUSB_ERROR_NO_DEVICE) {
                return -1;
            } else { //no such device
                int s = reconnect_usb();
                if(s < 0) return -1;
                continue; //retry
            }
        } else {
            check_to_receive_zero_byte(transfered, EP_IN_BULK);
            break;
        }
    }

    KnTsPrint(KN_LOG_DBG, "usb read received data :%d.\n", transfered);
    return transfered;
}

int USBDev::usb_recv(int8_t* msgbuf, int msglen) {
    if (NULL == m_pkneron_device_h) return -1;

    int transfered = 0;
    int ret_val = libusb_bulk_transfer(m_pkneron_device_h, EP_IN_INT,
                                       (unsigned char*)msgbuf, msglen, &transfered, LIB_USB_TIMEOUT);
    if (ret_val != 0) {
        if(ret_val == LIBUSB_ERROR_NO_DEVICE) {
            KnTsPrint(KN_LOG_ERR, "USB EP read failed:%d, %s.\n", ret_val, libusb_error_name(ret_val));
            reconnect_usb(); //reconnect to usb in case of no device
        }
        return ret_val;
    }

    check_to_receive_zero_byte(transfered, EP_IN_INT);

    kn_dump_msg(msgbuf, transfered, false);

    //unpack the data
    KnPrint(KN_LOG_DBG, "unpacking msg payload...\n");
    UARTPkt pkt(msgbuf, transfered);
    if(!pkt.is_pkt_valid()) {
        KnPrint(KN_LOG_ERR, "unpacking UART packet falied.\n");
        return -1;
    }

    int bytes_received = pkt.get_payload_len();
    assert(bytes_received <= msglen);
    memset(msgbuf, 0, msglen);
    memcpy(msgbuf, pkt.get_payload_buf(), bytes_received);

    KnPrint(KN_LOG_DBG, "get msg payload from uart pkt:%d.\n", bytes_received);

    //payload
    if(bytes_received < (int)sizeof(KnMsgHdr)) {  //less than header
        KnPrint(KN_LOG_ERR, "receive msg len less than hdr.\n");
    } else {
        KnMsgHdr hdr;
        KnBaseMsg::decode_msg_hdr(&hdr, msgbuf, bytes_received);
        int mlen = hdr.msg_len;
        if(bytes_received < mlen + (int)sizeof(KnMsgHdr)) { //only a part msg
            KnPrint(KN_LOG_ERR, "only a part of msg received:%d,%d.\n", bytes_received, mlen);
        }
        //whole msg is ready
    }

    KnPrint(KN_LOG_DBG, "from com dev, got %d bytes msg.\n", bytes_received);

    //kn_dump_msg(msgbuf, bytes_received, false);
    return bytes_received;
}


int USBDev::recv_data_ack() {
    int8_t ack_buffer[ACK_PACKET_SIZE];
    memset(ack_buffer, 0, sizeof(ack_buffer));
    //uint8_t usb_ack[8] = { 0x35, 0x8A, 0x04, 0x00, 0x04, 0x00, 0x08, 0x00 };

    KnTsPrint(KN_LOG_DBG, "usb dev is receiving ack...\n");
    int status = receive_bulk((char*)ack_buffer, ACK_PACKET_SIZE);
    if(status < 4) {
        KnPrint(KN_LOG_ERR, "receiving ack failed:%d.\n", status);
        return -1;
    }

    //get the ack msg
    KnImgTransAck ack_msg(ack_buffer+4, status-4);
    if(ack_msg.is_msg_valid() == false) {
        KnPrint(KN_LOG_ERR, "ack msg parsing failed:%d.\n", status);
        return -1;
    }

    if(ack_msg._action != 0) { //not ack
        KnPrint(KN_LOG_ERR, "got nack msg from device %d.\n", ack_msg._action);
        return -1;
    }

    return 0;
}

int USBDev::send_data_ack(uint32_t action, uint32_t src_id, uint32_t len) {
    uint8_t ack_buffer[ACK_PACKET_SIZE] = { 0x83, 0xA5, 0x04, 0x00, }; //pack header

    KnImgTransAck ack_msg(CMD_ACK_NACK, 8, action, src_id);
    if(ack_msg.is_msg_valid() == false) {
        KnPrint(KN_LOG_ERR, "creating ack msg failed.\n");
        return -1;
    }

    //packing msg
    memcpy(ack_buffer+4, ack_msg.get_msg_payload(), ack_msg.get_msg_len());
    memcpy(ack_buffer+12, &len, 4);

    //send it
    int status = send_bulk((char*)ack_buffer, ACK_PACKET_SIZE, true);
    if(status != ACK_PACKET_SIZE) {
        KnPrint(KN_LOG_ERR, "sending ack msg failed:%d.\n", status);
        return status;
    }

    KnTsPrint(KN_LOG_DBG, "sent ack msg to dev.\n");
    return 0;
}

int USBDev::send_image_data(char *buf, int data_len) {
    if(data_len <= 0 || !buf) return -1;

    KnTsPrint(KN_LOG_DBG, "usb dev is starting to send image :%d...\n", data_len);

    int status = recv_data_ack();
    if(status != 0) {
        KnPrint(KN_LOG_ERR, "ack msg wrong:%d.\n", status);
        return -1;
    }

    int leftover = data_len;
    int i = 0;
    int off = 0;

    while (leftover > 0) {
        int len = static_cast<int> (USB_MAX_BYTES);
        if (len > leftover) len = leftover;

        KnTsPrint(KN_LOG_DBG, "usb dev is sending segment :%d, with len:%d...\n", i, len);
        status = send_bulk(buf + off, len, false);
        if(status < 0) {
            KnPrint(KN_LOG_ERR, "sending %d time for img, off:%d.\n", i, off);
            return status;
        }

        KnTsPrint(KN_LOG_DBG, "usb dev sent segment :%d, with len:%d...\n", i, status);

        off += status;
        leftover -= status;
        i++;
    }

    check_to_send_zero_byte(data_len, EP_OUT_BULK);

    KnTsPrint(KN_LOG_DBG, "usb dev finished sending image :%d...\n", off);

    return off;
}

